from typing import List

import numpy as np
from qulacs import QuantumCircuit

from qulacsvis.utils.gate import grouping_adjacent_gates, to_latex_style
from qulacsvis.visualization.circuit_parser import (
    CircuitParser,
    ControlQubitInfo,
    GateData,
)


class LatexSourceGenerator:
    """Generate latex source from QuantumCircuit

    Parameters
    ----------
    circuit : QuantumCircuit
        A quantum circuit to be drawn.

    Attributes
    ----------
    _quantum_circuit : QuantumCircuit
        A quantum circuit to be drawn.
    _parser : CircuitParser
        The parser of the quantum circuit.
    _circuit_data : CircuitData
        The data of the quantum circuit after parsing by CircuitParser.
    _circuit : numpy.ndarray
        A matrix containing strings converted from CircuitData for Qcircuit.
        Each element and its position corresponds to one of GateData.
        Quantum circuit only, input values are not contained.
    _head : str
        The head of the latex source containing preamble.
    _tail : str
        The tail of the latex source.

    Examples
    --------
    >>> from qulacs import QuantumCircuit
    >>> from qulacsvis.visualization import LatexSourceGenerator
    >>>
    >>> circuit = QuantumCircuit(3)
    >>> circuit.add_X_gate(0)
    >>> circuit.add_Y_gate(1)
    >>> circuit.add_Z_gate(2)
    >>>
    >>> generator = LatexSourceGenerator(circuit)
    >>> latex_source = generator.generate()
    >>> print(latex_source)
    """

    def __init__(self, circuit: QuantumCircuit):
        self._quantum_circuit = circuit
        self._parser = CircuitParser(circuit)
        self._circuit_data = self._parser.parsed_circuit
        self._circuit = np.array([[]])
        self._head = r"""
\documentclass[border={-2pt 5pt 5pt -7pt}]{standalone}
\usepackage[braket, qm]{qcircuit}
\usepackage{graphicx}

\begin{document}
    \Qcircuit @C=1.0em @R=0.7em @!R{ \\
"""
        self._tail = r"    }" + "\n" + r"\end{document}"

    def generate(self) -> str:
        """Generate latex source from QuantumCircuit

        Returns
        -------
        latex_source : str
            String of latex source generated
        """
        qubit_count = self._parser.qubit_count
        circuit_layer_count = len(self._circuit_data[0])

        input_label = np.array(
            [
                [
                    # nghost reserves drawing area for input label,
                    # adjusts the spacing between rows.
                    r"\nghost{ q_{" + str(i) + "} : }",
                    r"\lstick{ q_{" + str(i) + "} : }",
                ]
                for i in range(qubit_count)
            ]
        )

        self._circuit = np.array([[] for _ in range(qubit_count)])
        for layer in range(circuit_layer_count):
            # This is an array containing strings.
            # The string for each element is the string for the Qcircuit of the gate
            # corresponding to each qubit row of the layer currently of interest.
            current_layer_latex = [to_latex_style("wire") for _ in range(qubit_count)]
            for qubit in range(qubit_count):
                gate = self._circuit_data[qubit][layer]

                if gate.name == "ghost":
                    continue
                elif gate.name == "wire":
                    continue
                elif gate.name == "CNOT":
                    self._cnot(current_layer_latex, gate)
                elif gate.name == "SWAP":
                    self._swap(current_layer_latex, gate)
                elif len(gate.target_bits) > 1:
                    self._multi_gate(current_layer_latex, gate)
                else:
                    self._gate(current_layer_latex, gate)

            self._circuit = np.column_stack([self._circuit, current_layer_latex])  # type: ignore
        wires = np.array([[r"\qw"] for _ in range(qubit_count)])
        circuit_with_label = np.column_stack([input_label, self._circuit, wires])  # type: ignore
        body = self._matrix_to_qcircuit_style(circuit_with_label)

        return self._head + body + self._tail

    def _matrix_to_qcircuit_style(self, matrix: List[List[str]]) -> str:
        """Generate from matrix to string for Qcircuit

        Parameters
        ----------
        matrix : List[List[str]]
            A matrix containing strings converted from CircuitData for Qcircuit.
        Returns
        -------
        res : str
            Generated string for Qcircuit from matrix
        """
        lines = [" & ".join(line) for line in matrix]
        # add indent for latex source file
        indent = "        "
        lines = [indent + line for line in lines]
        res = (r"\\" + "\n").join(lines)
        res += r"\\" + "\n"
        return res

    def _cnot(self, layer_latex: List[str], gate: GateData) -> None:
        """Generate CNOT gate for Qcircuit

        Parameters
        ----------
        layer_latex : List[str]
            Array containing the string of the gate of the layer currently of interest
        gate : GateData
            The gate data to be drawn.
        """
        cnot_qcircuit_style = to_latex_style(gate.name)
        target_bit = gate.target_bits[0]
        layer_latex[target_bit] = cnot_qcircuit_style
        self._control_bits(layer_latex, gate.control_bit_infos, target_bit)

    def _control_bits(
        self,
        layer_latex: List[str],
        control_bit_infos: List[ControlQubitInfo],
        target_bit: int,
    ) -> None:
        """Generate control bits for Qcircuit

        Parameters
        ----------
        layer_latex : List[str]
            Array containing the string of the gate of the layer currently of interest
        control_bits : List[int]
            The control bits of the gate to be drawn.
        target_bit : int
            A target bit of the gate.
            This value is used to generate the line connecting the target bit and control bits.
        """

        if len(control_bit_infos) == 0:
            return

        merged_bit_infos = [ControlQubitInfo(target_bit, 1)]
        merged_bit_infos.extend(control_bit_infos)
        merged_bit_infos.sort(key=lambda x: x.index)

        for info, next_info in zip(merged_bit_infos, merged_bit_infos[1:]):
            if info.index >= target_bit:
                info, next_info = next_info, info

            control_bit = info.index
            if info.control_value == 0:
                ctrl_cmd = r"\ctrlo{"
            else:
                ctrl_cmd = r"\ctrl{"
            layer_latex[control_bit] = (
                ctrl_cmd + str(next_info.index - info.index) + "}"
            )

    def _swap(self, layer_latex: List[str], gate: GateData) -> None:
        """Generate SWAP gate for Qcircuit

        Parameters
        ----------
        layer_latex : List[str]
            Array containing the string of the gate of the layer currently of interest
        gate : GateData
            The gate data to be drawn.
        """
        swap_qcircuit_style = to_latex_style(gate.name)
        target_index_list = gate.target_bits
        swap = (target_index_list[0], target_index_list[-1])
        layer_latex[swap[0]] = swap_qcircuit_style
        layer_latex[swap[0]] += r" \qwx[" + str(swap[1] - swap[0]) + r"]"
        layer_latex[swap[1]] = swap_qcircuit_style

    def _multi_gate(self, layer_latex: List[str], gate: GateData) -> None:
        """Generate multi gate for Qcircuit (e.g., DensityMatrixGate)

        Parameters
        ----------
        layer_latex : List[str]
            Array containing the string of the gate of the layer currently of interest
        gate : GateData
            The gate data to be drawn.
        """
        gate_name_qcircuit_style = to_latex_style(gate.name)
        groups_adjacent_gates = grouping_adjacent_gates(gate.target_bits)

        self._control_bits(layer_latex, gate.control_bit_infos, gate.target_bits[0])

        for adjacent_gates in groups_adjacent_gates:
            size = len(adjacent_gates) - 1
            target_bit = adjacent_gates[0]
            # The "\multi_gate" should be placed in one location and "\ghost" in the other.
            layer_latex[target_bit] = (
                r"\multigate{" + str(size) + "}{" + gate_name_qcircuit_style + "}"
            )

            for target_bit in adjacent_gates[1:]:
                layer_latex[target_bit] = r"\ghost{" + gate_name_qcircuit_style + "}"

        if len(groups_adjacent_gates) > 1:
            # Generate the line connecting the each group
            for group1, group2 in zip(groups_adjacent_gates, groups_adjacent_gates[1:]):
                from_ = group1[-1]
                to_ = group2[0]
                size = to_ - from_
                layer_latex[from_] += r" \qwx[" + str(size) + "]"

    def _gate(self, layer_latex: List[str], gate: GateData) -> None:
        """Generate a gate for Qcircuit

        Parameters
        ----------
        layer_latex : List[str]
            Array containing the string of the gate of the layer currently of interest
        gate : GateData
            The gate data to be drawn.
        """
        gate_qcircuit_style = r"\gate{" + to_latex_style(gate.name) + "}"
        target_bit = gate.target_bits[0]
        layer_latex[target_bit] = gate_qcircuit_style
        self._control_bits(layer_latex, gate.control_bit_infos, target_bit)



import dataclasses
from collections import deque
from typing import Deque, List

import dataclasses_json
from qulacs import QuantumCircuit

GATE_DEFAULT_WIDTH = 1.0
GATE_DEFAULT_HEIGHT = 1.5


@dataclasses_json.dataclass_json
@dataclasses.dataclass
class ControlQubitInfo:
    index: int
    control_value: int


@dataclasses_json.dataclass_json
@dataclasses.dataclass
class GateData:
    name: str
    target_bits: List[int] = dataclasses.field(default_factory=list)
    control_bit_infos: List[ControlQubitInfo] = dataclasses.field(default_factory=list)


CircuitData = List[List[GateData]]
QueueCircuitData = List[Deque[GateData]]


class CircuitParser:
    """
    Parse quantum circuit into a list of gate data.

    Parameters
    ----------
    circuit : QuantumCircuit
        Quantum circuit to be parsed.

    Attributes
    ----------
    qubit_count : int
        Number of qubits in the circuit.
    gate_info : CircuitData
        List of gate data.
    layer_width : List[float]
        Width of each layer.
    """

    def __init__(self, circuit: QuantumCircuit):
        self.qubit_count = circuit.get_qubit_count()
        self.layer_width: List[float] = []
        self.__temp_parsed_circuit: QueueCircuitData = [
            deque() for _ in range(self.qubit_count)
        ]
        self.parsed_circuit: CircuitData = [[]]

        for position in range(circuit.get_gate_count()):
            gate = circuit.get_gate(position)
            target_index_list = gate.get_target_index_list()
            control_index_list = gate.get_control_index_list()
            control_index_value_list = [
                ControlQubitInfo(index, control_value)
                for index, control_value in gate.get_control_index_value_list()
            ]
            gate_name = gate.get_name()

            if len(target_index_list) == 0:
                print(
                    f"WARNING: The {position}-th Gate you added is skipped."
                    'This gate does not have "target_qubit_list".'
                )
                continue

            merged_index = target_index_list + control_index_list
            min_merged_index = min(merged_index)
            max_merged_index = max(merged_index)
            self._align_layers(min_merged_index, max_merged_index)

            for index in range(min_merged_index, max_merged_index + 1):
                if index == target_index_list[0]:
                    self.__temp_parsed_circuit[index].append(
                        GateData(gate_name, target_index_list, control_index_value_list)
                    )
                elif index in target_index_list:
                    self.__temp_parsed_circuit[index].append(GateData("ghost"))
                else:
                    self.__temp_parsed_circuit[index].append(GateData("wire"))

        self._align_layers(0, self.qubit_count)
        self.parsed_circuit = [list(queue) for queue in self.__temp_parsed_circuit]
        self.layer_width = [
            GATE_DEFAULT_WIDTH for _ in range(len(self.parsed_circuit[0]))
        ]

    def _align_layers(self, min_line_index: int, max_line_index: int) -> None:
        """
        Align layer sizes for a specified range of rows.

        Parameters
        ----------
        min_line_index : int
            Minimum row index to be aligned.
        max_line_index : int
            Maximum row index to be aligned.
        """
        if min_line_index > max_line_index:
            min_line_index, max_line_index = max_line_index, min_line_index
        lines = self.__temp_parsed_circuit[min_line_index : max_line_index + 1]
        layer_counts = [len(queue) for queue in lines]
        max_layer_count = max(layer_counts)

        for queue, layer_count in zip(lines, layer_counts):
            for _ in range(max_layer_count - layer_count):
                queue.append(GateData("wire"))



from .circuit_drawer import circuit_drawer  # noqa
from .circuit_parser import CircuitParser  # noqa
from .latex import LatexSourceGenerator  # noqa
from .matplotlib import MPLCircuitlDrawer  # noqa
from .text import TextCircuitDrawer  # noqa
